#include <stdio.h>
#include <stddef.h>
#include <assert.h>
#include <stdlib.h>
#include <vector>
#include "ufbx.h"

void create_vertex_buffer(const void *vertices, size_t count)
{
    printf(".. %zu vertices\n", count);
}

void create_index_buffer(const uint32_t *indices, size_t count)
{
    printf(".. %zu indices\n", count);
}

// GPU vertex representation.
// In practice you would need to use more compact custom vector types as
// by default `ufbx_real` is 64 bits wide.
struct Vertex {
    ufbx_vec3 position;
    ufbx_vec3 normal;
    ufbx_vec2 uv;
};

void convert_mesh_part(ufbx_mesh *mesh, ufbx_mesh_part *part)
{
    std::vector<Vertex> vertices;
    std::vector<uint32_t> tri_indices;
    tri_indices.resize(mesh->max_face_triangles * 3);

    // Iterate over each face using the specific material.
    for (uint32_t face_index : part->face_indices) {
        ufbx_face face = mesh->faces[face_index];

        // Triangulate the face into `tri_indices[]`.
        uint32_t num_tris = ufbx_triangulate_face(
            tri_indices.data(), tri_indices.size(), mesh, face);

        // Iterate over each triangle corner contiguously.
        for (size_t i = 0; i < num_tris * 3; i++) {
            uint32_t index = tri_indices[i];

            Vertex v;
            v.position = mesh->vertex_position[index];
            v.normal = mesh->vertex_normal[index];
            v.uv = mesh->vertex_uv[index];
            vertices.push_back(v);
        }
    }

    // Should have written all the vertices.
    assert(vertices.size() == part->num_triangles * 3);

    // Generate the index buffer.
    ufbx_vertex_stream streams[1] = {
        { vertices.data(), vertices.size(), sizeof(Vertex) },
    };
    std::vector<uint32_t> indices;
    indices.resize(part->num_triangles * 3);

    // This call will deduplicate vertices, modifying the arrays passed in `streams[]`,
    // indices are written in `indices[]` and the number of unique vertices is returned.
    size_t num_vertices = ufbx_generate_indices(
        streams, 1, indices.data(), indices.size(), nullptr, nullptr);

    // Trim to only unique vertices.
    vertices.resize(num_vertices);

    create_vertex_buffer(vertices.data(), vertices.size());
    create_index_buffer(indices.data(), indices.size());
}

int main(int argc, char **argv)
{
    ufbx_scene *scene = ufbx_load_file("material_parts.fbx", NULL, NULL);
    assert(scene);

    for (ufbx_mesh *mesh : scene->meshes) {
        printf("mesh '%s'\n", mesh->name.data);

        for (ufbx_mesh_part &part : mesh->material_parts) {
            ufbx_material *material = NULL;
            if (part.index < mesh->materials.count) {
                material = mesh->materials.data[part.index];
            }
            printf(". [%u] material: %s\n", part.index,
                material ? material->name.data : "(none)");
            convert_mesh_part(mesh, &part);
        }
    }

    return 0;
}

